--
-- Scratchpad MMU - Core VHDL module
-- Copyright (C) Jack Whitham 2009-2010
-- http://www.jwhitham.org.uk/c/smmu.html
--
-- This library is free software; you can redistribute it and/or
-- modify it under the terms of the GNU Lesser General Public
-- License as published by the Free Software Foundation
-- (version 2.1 of the License only).
-- 
-- This library is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-- Lesser General Public License for more details.
-- 
-- You should have received a copy of the GNU Lesser General Public
-- License along with this library; if not, write to the Free Software
-- Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
-- 

library ieee;
library unisim ;

use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned."+";
use ieee.std_logic_unsigned."-";
use ieee.std_logic_unsigned.conv_integer;
use unisim.vcomponents.all ;


entity smmu_core is
    -- Configuration
    generic (
        NUM_READ_ONLY_ENTRIES : Natural; 
        BURST_SIZE_LEFT             : Natural := 3;
        INITIATE_POINT              : Natural := 1;
        NUM_ENTRIES                 : Natural;
        LOG2_NUM_CACHE_LINES_PER_PAGE : Natural := 0;
        LOG2_NUM_CACHE_WORDS_PER_LINE : Natural := 0;
        CPU_LEFT                    : Natural ;
        SP_LEFT                     : Natural ;
        DATA_LEFT                   : Natural ;
        REF_LEFT                    : Natural ;
        ACTIVE_LEFT                 : Natural ) ;
    port (
        -- CPU side - really these are external ports
        cpu_address         : in std_logic_vector ( CPU_LEFT downto 0 );
            -- also used for OPEN (base address), cache invalidate and
            -- permutation programming
        cpu_data_in         : in std_logic_vector ( DATA_LEFT downto 0 );
            -- cpu_data_in is also used for "size" and "reference in"
            -- and cache permutation input
        cpu_data_out        : out std_logic_vector ( DATA_LEFT downto 0 );
            -- cpu_data_out is also used for "reference out"
        cpu_copy_loc        : in std_logic_vector ( SP_LEFT downto 0 );
        cpu_size            : in std_logic_vector ( SP_LEFT downto 0 );
        cpu_read_only       : in std_logic;
        cpu_ben             : in std_logic_vector ( 3 downto 0 );
        cpu_write           : in std_logic; -- in 
        cpu_read            : in std_logic; -- in 
        cpu_open            : in std_logic; -- in 
        cpu_close           : in std_logic; -- in 
        cpu_ack             : out std_logic; -- out 
        -- Add Data LMB interface here
        -- Control lines from smmu to IPIF
        mst_cntl_rd_req     : out std_logic ;
        mst_cntl_wr_req     : out std_logic ;
        mst_cntl_bus_lock   : out std_logic ;
        mst_cntl_burst      : out std_logic ;
        mst_ip2bus_addr     : out std_logic_vector ( CPU_LEFT downto 0 ) ;
        mst_ip2bus_be       : out std_logic_vector ( 15 downto 0 ) ;
        mst_xfer_length     : out std_logic_vector ( 11 downto 0 ) ;
        mst_go              : out std_logic ;
        -- Status lines from IPIF to smmu
        mst_cmd_sm_set_done : in std_logic ;
        mst_cmd_sm_set_error : in std_logic ;
        mst_cmd_sm_set_timeout : in std_logic ;
        mst_cmd_sm_busy     : in std_logic ;
        mst_cmd_sm_clr_go   : in std_logic ;
        -- Data interface
        output_next         : in std_logic ;
        input_next          : in std_logic ;
        data_output         : out std_logic_vector ( DATA_LEFT downto 0 ) ;
        data_input          : in std_logic_vector ( DATA_LEFT downto 0 ) ;
        -- Only service addresses matching this:
        match_mask          : in std_logic_vector ( CPU_LEFT 
                                downto ACTIVE_LEFT + 1 ) ;
        -- Cache controls
        cache_active        : in std_logic;
        cpu_cache_set_perm  : in std_logic;
        cpu_cache_invalidate: in std_logic;
        -- DEBUGGING HARNESS ONLY
    dbg_cpu_side_address    : out std_logic_vector ( SP_LEFT downto 0 ) ;
    dbg_bus_side_address    : out std_logic_vector ( SP_LEFT downto 0 ) ;
    dbg_cpu_side_data_in    : out std_logic_vector ( DATA_LEFT downto 0 ) ;
    dbg_cpu_side_data_out   : out std_logic_vector ( DATA_LEFT downto 0 ) ;
    dbg_bus_side_data_in    : out std_logic_vector ( DATA_LEFT downto 0 ) ;
    dbg_bus_side_data_out   : out std_logic_vector ( DATA_LEFT downto 0 ) ;
    dbg_cpu_side_write      : out std_logic ;
    dbg_bus_side_write      : out std_logic ;
    dbg_bus_side_cen        : out std_logic ;
    dbg_match_address       : out std_logic_vector ( CPU_LEFT downto 0 ) ;
    dbg_match               : out std_logic;
    dbg_match_in_read_only_area: out std_logic;
    dbg_miss                : out std_logic;
    dbg_hit                 : out std_logic;
    dbg_read_only_update_pos  : out std_logic_vector ( REF_LEFT downto 0 );
    dbg_update_burst_boundary : out std_logic;
    dbg_code                : out std_logic_vector ( 7 downto 0 );
    dbg_sp_cpu_valid        : out std_logic;
    dbg_slow_ack            : out std_logic;
    dbg_row_activation      : out std_logic_vector ( NUM_ENTRIES - 1 downto 0 );
    dbg_match_ref           : out std_logic_vector ( REF_LEFT downto 0 );
    dbg_tag_address         : out std_logic_vector ( SP_LEFT - 1 downto 0 );
    dbg_permutation_in      : out std_logic_vector ( SP_LEFT - 1 downto 0 );
    dbg_cache_hit           : out std_logic;
    dbg_tag_in_16           : out std_logic_vector ( 15 downto 0 );
    dbg_tag_out_16          : out std_logic_vector ( 15 downto 0 );
    dbg_permutation_out     : out std_logic_vector ( 15 downto 0 );
    dbg_cache_read          : out std_logic;
    dbg_cache_write         : out std_logic;
    dbg_enable_permutation  : out std_logic;
        -- Clock
        clk                 : in std_logic ;
        reset               : in std_logic ) ;
end entity smmu_core ;


architecture basic of smmu_core is

    -- Cache feature - enabled?
    function Max(a, b: Natural) return Natural is
    begin
        if (a > b) then return a; 
        else return b; 
        end if;
    end function Max;

    constant CACHE_FEATURE : Boolean := ( LOG2_NUM_CACHE_WORDS_PER_LINE /= 0 ) ;
    constant CACHE_PAGE_BITS : Natural := Max(LOG2_NUM_CACHE_LINES_PER_PAGE, 1);
    constant CACHE_LINE_BITS : Natural := Max(LOG2_NUM_CACHE_WORDS_PER_LINE, 1);

    -- Bus types
    subtype Bus_CPU_Data is std_logic_vector ( DATA_LEFT downto 0 ) ;
    subtype Bus_CPU_Address is std_logic_vector ( CPU_LEFT downto 0 ) ;
    subtype Bus_Match_Address is std_logic_vector ( ACTIVE_LEFT downto 0 ) ;
    subtype Bus_SP_Address is std_logic_vector ( SP_LEFT downto 0 ) ;

    signal master_ack       : std_logic; 

    -- Internal control lines from state machine
    signal open_operation   : std_logic;
    signal close_operation  : std_logic; 
    signal trigger          : std_logic; 
    signal next_ack         : std_logic;
    signal sp_write         : std_logic; 
    signal mst_go_int       : std_logic;
    signal bus_side_cen_override: std_logic;
    signal match_internal   :std_logic;
    signal read_only_copy_enable:std_logic;
    signal cache_read       :std_logic;
    signal cache_write      :std_logic;
    signal cache_read_d     :std_logic;
    signal cache_write_d    :std_logic;

    -- Other control lines
    signal sp_cpu_valid     : std_logic; -- delayed hit
    signal cpu_side_write   : std_logic_vector ( 3 downto 0 ) ;
    signal bus_side_write   : std_logic_vector ( 3 downto 0 ) ;
    signal match_delay      : std_logic; -- delayed match
    signal match_inhibit    : std_logic;
    signal bus_side_cen     : std_logic;

    -- Status lines
    signal hit              : std_logic;
    signal miss             : std_logic;
    signal busy             : std_logic;
    signal some_words_remain: std_logic;
    signal one_word_remains : std_logic;
    signal close_null       : std_logic;
    signal cache_write_hit  : std_logic;
    signal cache_read_hit   : std_logic;
    signal ep_trap          : std_logic;
    signal port_b_write     : std_logic;

    -- Internal buses
    signal cpu_side_data_out: Bus_CPU_Data ;
    signal slow_data_out    : Bus_CPU_Data ;
    signal cpu_side_data_in : Bus_CPU_Data ;
    signal bus_side_data_out: Bus_CPU_Data ;
    signal bus_side_data_in : Bus_CPU_Data ;
    signal match_in_address : Bus_Match_Address ;
    signal match_out_address: Bus_SP_Address ;
    signal close_start_address : Bus_SP_Address;
    signal entry_size       : Bus_SP_Address;
    signal port_b_address   : Bus_SP_Address ;
    signal bus_side_address_inc : Bus_SP_Address ;

    -- Internal registers
    signal data_input_delay : Bus_CPU_Data ;
    signal bus_side_address : Bus_SP_Address ;
    signal words_remaining  : Bus_SP_Address ;
    signal match_address    : Bus_CPU_Address ;
    signal cache_address    : Bus_SP_Address ;
    signal saved_data_in    : Bus_CPU_Data ;
    signal slow_ack         : std_logic; 
    signal saved_be         : std_logic_vector ( 3 downto 0 ) ;
    signal saved_read_only  : std_logic;
    signal cache_access     : std_logic;
    signal master_address   : std_logic_vector ( CPU_LEFT downto 0 ) ;
    signal cache_saved_address: 
                        std_logic_vector ( SP_LEFT downto 0 ) ;

    -- The table

    signal ref              : Natural range 0 to NUM_ENTRIES - 1;
    signal read_only_update_pos: Natural range 0 to NUM_READ_ONLY_ENTRIES;
    signal next_update_pos  : Natural range 0 to NUM_READ_ONLY_ENTRIES ;

    type TABLE_ENTRY is record
                is_valid            : std_logic ;
                min_address         : Bus_Match_Address ;
                max_address         : Bus_Match_Address ;
                min_address_short   : Bus_SP_Address ;
                max_address_short   : Bus_SP_Address ;
                offset              : Bus_SP_Address ;
                --out_address         : Bus_SP_Address ;
            end record ;

    type TABLE_ARRAY is array (0 to NUM_ENTRIES-1) of TABLE_ENTRY;

    signal table            : TABLE_ARRAY ;

    -- The state machine
    type State_Type is ( 
            READY , SLOW_ACCESS_WAIT , SLOW_RO_UPDATE,
            BEGIN_OPEN , STAGE1_OPEN , STAGE2_OPEN , STAGE3_OPEN ,
            BEGIN_CLOSE , STAGE1_CLOSE , STAGE2_CLOSE , STAGE3_CLOSE ,
            SLOW_RO_UPDATE_1, STAGE3A_CLOSE, CACHE_MISS_WAIT,
            CACHE_MISS_WAIT_ONE, CACHE_MISS,
            HANDLE_TABLE_FULL , HANDLE_ERROR , HANDLE_TIMEOUT ,
            HANDLE_NOT_FOUND, HANDLE_DEADLINE_MISS) ;
    
    signal state            : State_Type ;

    -- Matching
    signal match_ro         : std_logic;
    signal match_rw         : std_logic;
    signal row_match        : std_logic_vector ( 0 to NUM_ENTRIES - 1 ) ;
    signal row_activation   : std_logic_vector ( 0 to NUM_ENTRIES - 1 ) ;
    signal row_validity     : std_logic_vector ( 0 to NUM_ENTRIES - 1 ) ;
    signal row_match_1      : std_logic_vector ( 0 to NUM_ENTRIES - 1 ) ;
    signal row_match_2      : std_logic_vector ( 0 to NUM_ENTRIES - 1 ) ;
    signal shift_reg_out    : std_logic_vector ( DATA_LEFT downto 0 ) ;

    constant BURST_SIZE : Natural := ( 2 ** ( BURST_SIZE_LEFT + 1 )) ;

    signal close_update_count : Natural range 0 to BURST_SIZE - 1 ;

    signal min_address_cat  : std_logic_vector(
                        ((ACTIVE_LEFT + 1) * NUM_ENTRIES) - 1 downto 0);
    signal max_address_cat  : std_logic_vector(
                        ((ACTIVE_LEFT + 1) * NUM_ENTRIES) - 1 downto 0);
    signal out_address_cat  : std_logic_vector(
                        ((SP_LEFT + 1) * NUM_ENTRIES) - 1 downto 0);

begin 

    assert NUM_ENTRIES = 2 ** ( REF_LEFT + 1 ) ;

    -- First stage: comparator network identifies "hits" on the SMMU
    cn : entity comp_network 
        generic map ( 
            ACTIVE_LEFT => ACTIVE_LEFT ,
            NUM_ENTRIES => NUM_ENTRIES )
        port map (
            match_in_address => match_in_address,
            min_address_cat => min_address_cat,
            max_address_cat => max_address_cat,
            row_match_1 => row_match_1,
            row_match_2 => row_match_2 );

    -- Second stage: priority encoder identifies the highest
    -- priority row with a match
    pd : entity prio_dec 
        generic map ( 
            NUM_READ_ONLY_ENTRIES => NUM_READ_ONLY_ENTRIES,
            REF_LEFT => REF_LEFT,
            SP_LEFT => SP_LEFT,
            NUM_ENTRIES => NUM_ENTRIES )
        port map (
            out_address_cat => out_address_cat,
            match_out_address => match_out_address,
            match_ro => match_ro,
            match_rw => match_rw,
            row_activation => row_activation,
            row_match_1 => row_match_1,
            row_match_2 => row_match_2 );


    process ( match_in_address , table ) is
        variable x : Natural;
    begin
        out_address_cat <= (others => '0');
        for i in 1 to ( NUM_ENTRIES - 1 )
        loop
            -- out_address computation
            x := i * (SP_LEFT + 1);
            out_address_cat(x + SP_LEFT downto x) <=
                    match_in_address ( SP_LEFT downto 0 ) + 
                    table(i).offset;
        end loop ;
    end process ;

    process ( match_internal , match_mask ,
                cpu_address , match_address ) is
    begin
        -- match_in_address multiplexer
        match_inhibit <= '0' ;
        if ( match_internal = '1' )
        then
            -- OPEN/CLOSE operation
            match_in_address <= match_address ( ACTIVE_LEFT downto 2 ) ;
        else
            -- Regular access from CPU
            match_in_address <= cpu_address ( ACTIVE_LEFT downto 2 ) ;
            if ( cpu_address ( 
                    CPU_LEFT downto ACTIVE_LEFT + 1 ) /= match_mask )
            then
                match_inhibit <= '1' ;
            end if ;
        end if ;
    end process;

    -- "Hits" are acknowledged on the following clock cycle.
    -- "Misses" are serviced by the memory access mechanism.
    hit <= cpu_write or cpu_read when (match_ro or match_rw) = '1'
        else '0';
    miss <= cpu_read or cpu_write when (match_ro or match_rw) = '0'
        else cpu_write when match_ro = '1'
        else '0';
    master_ack <= sp_cpu_valid or slow_ack or cache_read_hit;
    cpu_ack <= master_ack;
    cpu_data_out <= cpu_side_data_out when ( sp_cpu_valid = '1' )
        else bus_side_data_out when ( cache_read_hit = '1' )
        else slow_data_out;
    cpu_side_data_in <= cpu_data_in when ( hit = '1' )
        else bus_side_data_out ;
    bus_side_data_in <= 
        data_input_delay when ((( open_operation and not match_delay ) = '1' )
                    or ( cache_write_hit = '1' ))
        else cpu_side_data_out ;
    busy <= mst_cmd_sm_busy or mst_go_int or trigger ;
    some_words_remain <= '1' when ( conv_integer ( words_remaining ) /= 0 )
        else '0' ;
    one_word_remains <= '1' when ( conv_integer ( words_remaining ) = 1 )
        else '0' ;
    data_output <= shift_reg_out when ( close_operation = '1' )
        else saved_data_in ;
    bus_side_cen <= bus_side_cen_override;

    -- Outgoing controls to PLB IPIF
    mst_cntl_bus_lock <= '0' ;
    mst_go <= mst_go_int ;
    mst_xfer_length <= 
        -- 1 word if a "slow" operation:
        x"004" when (( open_operation or close_operation ) = '0' )
        -- -- 1 word if words_remaining = 0
        -- else BURST_SIZE words if words_remaining >= BURST_SIZE
        else conv_std_logic_vector ( 0 , ( 12 - 4 ) - BURST_SIZE_LEFT ) 
            & "1" & conv_std_logic_vector ( 0 , BURST_SIZE_LEFT + 3 )
                when 0 /= conv_integer ( 
                    words_remaining ( SP_LEFT downto BURST_SIZE_LEFT + 1 )) 
        -- otherwise, words_remaining 
        else conv_std_logic_vector ( 0 , ( 12 - 3 ) - BURST_SIZE_LEFT ) 
                & words_remaining ( BURST_SIZE_LEFT downto 0 ) & "00" ;

    mst_cntl_burst <= ( open_operation or close_operation ) 
            and some_words_remain and not one_word_remains ;


    mst_ip2bus_addr <= master_address ;

    close_start_address <= 
        table(ref).offset + table(ref).min_address_short;

    entry_size <= 
        table(ref).max_address_short - table(ref).min_address_short;

    close_null <= close_operation when match_address ( SP_LEFT + 2 downto 2 ) 
                    = table(ref).max_address_short ( SP_LEFT downto 0 )
        else '0' ;

    lane_gen : for lane in 0 to 3
    generate 
        lane_block : block
            constant right  : Natural := 8 * lane ;
            constant left   : Natural := right + 7 ;
        begin
            -- Control for this byte lane
            cpu_side_write ( lane ) <= ( 
                ( read_only_copy_enable and not close_null )
                or ( cpu_write and cpu_ben ( lane ))) 
                    when (match_ro or match_rw) = '1'
                        else '0';
            bus_side_write ( lane ) <= saved_be ( lane ) and port_b_write;

            copy : for i in 0 to 3
            generate
                mst_ip2bus_be ( lane + ( i * 4 )) <= saved_be ( lane ) ;
            end generate copy ;

            -- Memory for this byte lane
            scratchpad : RAMB16_S9_S9
                port map (
                    addrA => match_out_address,
                    diA => cpu_side_data_in ( left downto right ),
                    dipA => "0",
                    doA => cpu_side_data_out ( left downto right ),
                    dopA => open,
                    clkA => clk,
                    enA => '1',
                    ssrA => reset,
                    weA => cpu_side_write ( lane ),

                    addrB => port_b_address,
                    diB => bus_side_data_in ( left downto right ),
                    dipB => "0",
                    doB => bus_side_data_out ( left downto right ),
                    dopB => open,
                    clkB => clk,
                    enB => bus_side_cen,
                    ssrB => reset,
                    weB => bus_side_write ( lane ));

        end block lane_block ;
    end generate lane_gen ;


    cache_no_lookup : if not CACHE_FEATURE
    generate
        port_b_write <= sp_write ;
        port_b_address <= bus_side_address;
        cache_write_hit <= '0';
        cache_read_hit <= '0';
    end generate cache_no_lookup;


    cache_lookup : if CACHE_FEATURE
    generate
        -- Typical configuration:
        -- ADDRESS BITS    PURPOSE
        --  31 to 24     inactive bits
        --  23 to 13     tag bits                             HIGH_BITS*
        --  12 to 8      tag bits & permutation (page select) LOG2_NUM_PAGES*
        --   7 to 5      line number within page              CACHE_PAGE_BITS
        --   4 to 2      word number                          CACHE_LINE_BITS
        --   1 to 0      byte number                          2
        -- * = derived 
        -- Note: 4 = CACHE_LINE_BITS + 1
        --       7 = CACHE_LINE_BITS + CACHE_PAGE_BITS + 1 = CACHE_X - 1
        --      12 = SP_LEFT + 2
        --      23 = ACTIVE_LEFT
        -- 

        cache_block : block
            constant LOG2_NUM_PAGES : Natural :=
                    1 + SP_LEFT - (CACHE_LINE_BITS + CACHE_PAGE_BITS);
            constant HIGH_BITS : Natural :=
                    ACTIVE_LEFT - (SP_LEFT + 2);

            constant PERM_BITS  : Natural := LOG2_NUM_PAGES;
            constant PERM_LEFT  : Natural := PERM_BITS - 1;
            constant PERM_SIZE  : Natural := 2 ** PERM_BITS;
            constant B18_LEFT   : Natural := SP_LEFT - 1;

            constant TAG_BITS   : Natural := LOG2_NUM_PAGES + HIGH_BITS;
            constant TAG_LEFT   : Natural := TAG_BITS - 1;
            constant CACHE_X    : Natural :=
                        CACHE_PAGE_BITS + CACHE_LINE_BITS + 2 ;

            signal permutation_in  : std_logic_vector ( PERM_LEFT downto 0 ) ;
            signal permutation_out : std_logic_vector ( PERM_LEFT downto 0 ) ;
            signal cpu_address_tag : std_logic_vector ( TAG_LEFT downto 0 ) ;
            signal tag_out         : std_logic_vector ( TAG_LEFT downto 0 ) ;
            signal tag_address     : std_logic_vector ( B18_LEFT downto 0 ) ;
            signal tag_out_18      : std_logic_vector ( 17 downto 0 ) ;
            signal tag_in_18       : std_logic_vector ( 17 downto 0 ) ;
            signal set_perm_data_16: std_logic_vector ( 15 downto 0 ) ;
            signal tag_out_valid   : std_logic;
            signal tag_write       : std_logic;
            signal invalidate      : std_logic;

            type PERM_RAM is array (0 to PERM_SIZE-1) of 
                        std_logic_vector ( PERM_LEFT downto 0 ) ;
            signal permutation_ram : PERM_RAM;
        begin
            -- Only 17 bits are available for tag space.
            -- If this is not enough, change the page size, line size,
            -- or the number of active bits.
            assert TAG_BITS < 18;
            assert PERM_BITS > 0;
            assert ( CACHE_LINE_BITS + CACHE_PAGE_BITS + 
                    LOG2_NUM_PAGES ) = SP_LEFT + 2 - 1;
            assert ( CACHE_LINE_BITS + CACHE_PAGE_BITS + LOG2_NUM_PAGES +
                    HIGH_BITS ) = ACTIVE_LEFT - 1;

            -- Wiring for cache units
            process ( cpu_address, permutation_out, master_address ,
                    cpu_address_tag, tag_out_18, sp_write ) is
            begin
                permutation_in <= cpu_address ( 
                            PERM_LEFT + CACHE_X downto CACHE_X ) ;
                cpu_address_tag <= 
                            master_address ( ACTIVE_LEFT downto CACHE_X ) ;

                tag_in_18 <= ( others => '0' ) ;
                tag_in_18 ( 17 ) <= sp_write;
                tag_in_18 ( TAG_LEFT downto 0 ) <= cpu_address_tag ;
                tag_out <= tag_out_18 ( TAG_LEFT downto 0 ) ;
                tag_out_valid <= tag_out_18 ( 17 ) ;

                tag_address <= ( others => '0' );
                tag_address ( PERM_LEFT + CACHE_PAGE_BITS downto 0 ) <= 
                        port_b_address ( CACHE_PAGE_BITS + PERM_LEFT +
                            CACHE_LINE_BITS downto CACHE_LINE_BITS );

                cache_address <= permutation_out ( PERM_LEFT downto 0 )
                        & cpu_address ( CACHE_PAGE_BITS +
                                CACHE_LINE_BITS + 2 - 1 downto 2 ) ;

                set_perm_data_16 <= ( others => '0' ) ;
                set_perm_data_16 ( PERM_LEFT downto 0 ) <= 
                        cpu_data_in ( PERM_LEFT downto 0 ) ;

            end process;

            process ( clk ) is
            begin
                if ( clk'event )
                and ( clk = '1' )
                then
                    dbg_permutation_in <= permutation_in;
                    dbg_enable_permutation <= '0';
                end if;
            end process;
            process ( clk, reset ) is
            begin
                if ( clk'event )
                and ( clk = '1' )
                then
                    dbg_tag_address <= tag_address;
                    invalidate <= cpu_cache_invalidate;
                end if;
            end process;

            dbg_permutation_out <= permutation_out;

            port_b_write <= cache_write_hit or sp_write;
            port_b_address <= 
                    cache_saved_address when (
                                cache_read_hit or cache_write_hit) = '1'
                    else bus_side_address when (state /= READY)
                    else cache_address;

            dbg_cache_read <= cache_read;
            dbg_cache_write <= cache_write;
            dbg_tag_in_16 <= tag_in_18(17 downto 2);
            dbg_tag_out_16 <= tag_out_18(17 downto 2);
            dbg_cache_hit <= cache_read_hit or cache_write_hit;

            -- Output valid in cycle 1
            tag_mem : RAMB16_S18
                port map (
                    -- Used for all tag accesses: read/invalidate/validate
                    addr => tag_address ,
                    di => tag_in_18 (15 downto 0),
                    dip => tag_in_18 (17 downto 16),
                    do => tag_out_18 (15 downto 0),
                    dop => tag_out_18 (17 downto 16),
                    clk => clk,
                    en => '1',
                    ssr => reset,
                    we => tag_write);

            permutation_out <= permutation_ram ( 
                    conv_integer ( permutation_in ));

            process (clk) is
            begin
                if ((clk = '1') and clk'event)
                then
                    if (cpu_cache_set_perm = '1')
                    then
                        permutation_ram ( conv_integer ( permutation_in ) )
                                <= set_perm_data_16 ( PERM_LEFT downto 0 );
                    end if;
                end if;
            end process;
                    

            tag_write <= sp_write or invalidate;

            process (cpu_address_tag, sp_cpu_valid,  tag_out,
                    cache_read_d, cache_write_d,
                    cache_active, cache_access) is
            begin
                cache_write_hit <= '0';
                cache_read_hit <= '0';
                if ((sp_cpu_valid = '0') and (cache_active = '1') 
                and (tag_out_valid = '1')
                and (tag_out = cpu_address_tag))
                then
                    cache_write_hit <= cache_write_d ;
                    cache_read_hit <= cache_read_d;
                end if;
            end process;

        end block cache_block ;
    end generate cache_lookup;

    dbg_row_activation <= row_activation ;
    dbg_sp_cpu_valid <= sp_cpu_valid;
    dbg_slow_ack <= slow_ack;
    dbg_match <= match_ro or match_rw;
    dbg_match_address <= match_address;
    dbg_miss <= miss;
    dbg_hit <= hit;
    dbg_cpu_side_data_out <= cpu_side_data_out;
    dbg_bus_side_data_out <= bus_side_data_out;
    dbg_bus_side_cen <= bus_side_cen;
    dbg_read_only_update_pos <= conv_std_logic_vector (
                read_only_update_pos , REF_LEFT + 1 ) ;
    dbg_update_burst_boundary <= close_null;


    bus_side_address_inc <= bus_side_address + 1 ;

    process ( table ) is
        variable x, y : Natural;
    begin
        for i in 0 to NUM_ENTRIES - 1
        loop
            y := i * (ACTIVE_LEFT + 1);
            x := y + ACTIVE_LEFT;
            row_validity(i) <= table(i).is_valid;
            table(i).min_address_short <= 
                        table(i).min_address ( SP_LEFT downto 0 ) ;
            table(i).max_address_short <= 
                        table(i).max_address ( SP_LEFT downto 0 ) ;
            min_address_cat(x downto y) <= 
                        table(i).min_address ( ACTIVE_LEFT downto 0 ) ;
            max_address_cat(x downto y) <= 
                        table(i).max_address ( ACTIVE_LEFT downto 0 ) ;
        end loop;
    end process;

    next_update_pos <= read_only_update_pos + 1;

    cache_write <= cpu_write and not match_inhibit;
    cache_read <= cpu_read and not match_inhibit;

    -- State machine
    process ( clk , reset , row_validity ) is
    begin
        if ( reset = '1' )
        then
            state <= READY;
            for i in 0 to NUM_ENTRIES - 1
            loop
                table(i).is_valid <= '0' ;
                table(i).min_address <= ( others => '0' ) ;
                table(i).max_address <= ( others => '0' ) ;
                table(i).offset <= ( others => '0' ) ;
            end loop;
            mst_go_int <= '0' ;
            sp_cpu_valid <= '0' ;
            match_internal <= '0';
            read_only_update_pos <= 1;
            slow_data_out <= x"e0e0e000" ;
            row_activation <= row_validity;
            dbg_code <= x"ee";

        elsif ( clk'event )
        and ( clk = '1' )
        then
            dbg_bus_side_address <= port_b_address;
            dbg_cpu_side_address <= match_out_address;
            dbg_cpu_side_data_in <= cpu_side_data_in;
            dbg_cpu_side_write <= cpu_side_write ( 0 );
            dbg_bus_side_data_in <= bus_side_data_in ;
            dbg_bus_side_write <= port_b_write;


            cache_read_d <= cache_read;
            cache_write_d <= cache_write;
            slow_ack <= next_ack ;
            next_ack <= '0' ;
            trigger <= '0' ;
            match_delay <= '0';
            sp_cpu_valid <= hit and not miss;
            sp_write <= '0' ;
            bus_side_cen_override <= '0' ;

            case state is
            when READY =>
                    match_internal <= '0';
                    read_only_copy_enable <= '0';
                    mst_cntl_rd_req <= '0' ;
                    mst_cntl_wr_req <= '0' ;
                    open_operation <= '0' ;
                    close_operation <= '0' ;
                    bus_side_cen_override <= '1' ;
                    match_address <= cpu_address ;
                    words_remaining <= cpu_size ;
                    saved_read_only <= cpu_read_only ;
                    read_only_update_pos <= 1;
                    ref <= conv_integer ( 
                            cpu_data_in ( REF_LEFT downto 0 ) ) ;
                    row_activation <= row_validity;
                    bus_side_address <= match_out_address;


                    if ( cpu_open = '1' )
                    then
                        state <= BEGIN_OPEN ;
                        --bus_side_address <= match_out_address;
                        master_address <= cpu_address ;
                        saved_data_in <= cpu_data_in ;
                        saved_be <= x"f" ;
                    elsif ( cpu_close = '1' )
                    then
                        state <= BEGIN_CLOSE ;
                        master_address <= cpu_address ;
                        saved_data_in <= cpu_data_in ;
                        saved_be <= x"f" ;
                    elsif ( miss = '1' )
                    then
                        -- Slow load or slow store; service
                        -- through main memory. Set up the 
                        -- transaction now.
                        mst_cntl_rd_req <= not cpu_write ;
                        mst_cntl_wr_req <= cpu_write ;
                        if ( hit = '1' )
                        then
                            -- (hit and miss) = '1' only when it's a write
                            -- within a read-only area.
                            match_internal <= '1';
                            -- Load new word into bus_side_data_out
                            bus_side_cen_override <= '1';
                            state <= SLOW_RO_UPDATE_1 ;
                            trigger <= '1' ;
                            saved_data_in <= cpu_data_in ;
                        elsif (( cache_active = '0' )
                        or not CACHE_FEATURE
                        or ( match_inhibit = '1' ))
                        then
                            trigger <= '1' ;
                            dbg_code <= x"22";
                            if ( match_inhibit = '1' )
                            then
                                dbg_code <= x"2f";
                            end if;
                            state <= SLOW_ACCESS_WAIT ;
                            saved_data_in <= cpu_data_in ;
                        end if;
                        master_address <= cpu_address ;
                        saved_be <= cpu_ben ;

                    elsif (( cache_write_d = '1' )
                    and CACHE_FEATURE
                    and ( cache_active = '1' ))
                    then
                        -- Cache write through
                        mst_cntl_rd_req <= '0';
                        mst_cntl_wr_req <= '1';
                        trigger <= '1' ;
                        state <= SLOW_ACCESS_WAIT ;
                        dbg_code <= x"41";

                    elsif (( cache_read_d = '1' )
                    and CACHE_FEATURE
                    and ( cache_read_hit = '0' )
                    and ( cache_active = '1' ))
                    then
                        -- Second cycle of cache access is a MISS
                        -- (and a read) Oh no!
                        state <= CACHE_MISS;
                    end if;
                    -- First cycle of cache access:- 
                    if (CACHE_FEATURE
                    and (((( cpu_write or cpu_read) = '1')
                            and ( cache_active = '1' ))
                        or (( cpu_cache_set_perm
                            or cpu_cache_invalidate ) = '1'))
                    and ( match_inhibit = '0' ))
                    then
                        master_address <= cpu_address ;
                        dbg_code ( 7 downto 4 ) <= x"3";
                        next_ack <= cpu_cache_invalidate
                                or cpu_cache_set_perm;
                        cache_saved_address <= cache_address;
                        saved_data_in <= cpu_data_in ;
                        data_input_delay <= cpu_data_in ;
                        bus_side_cen_override <= '1' ;
                        saved_be <= cpu_ben ;
                    end if;


            when SLOW_RO_UPDATE_1 =>
                    -- Process write to read-only area
                    row_activation <= (others => '0');
                    row_activation(1) <= row_validity(1);
                    read_only_copy_enable <= '1';
                    state <= SLOW_RO_UPDATE ;

            when SLOW_RO_UPDATE =>
                    -- The purpose of this state is (1) to wait for
                    -- a write to external memory to complete, and
                    -- (2) to update read-only areas as appropriate.
                    if ( read_only_update_pos /= NUM_READ_ONLY_ENTRIES )
                    then
                        row_activation <= (others => '0');
                        row_activation(next_update_pos) <= 
                                row_validity(next_update_pos);
                        read_only_update_pos <= next_update_pos;
                    else
                        read_only_copy_enable <= '0';
                    end if;
                    if (( busy = '0' )
                    and ( read_only_update_pos = NUM_READ_ONLY_ENTRIES ))
                    then
                        next_ack <= '1' ;
                        state <= READY ;
                    end if ;

            when SLOW_ACCESS_WAIT =>
                    if ( input_next = '1' )
                    then
                        slow_data_out <= data_input ;
                    end if ;
                    if ( busy = '0' )
                    then
                        next_ack <= '1' ;
                        state <= READY ;
                        dbg_code <= x"26";
                    end if ;

            when CACHE_MISS =>
                    bus_side_address <= cache_saved_address;
                    bus_side_address ( 
                                CACHE_LINE_BITS - 1 downto 0 ) 
                                <= ( others => '0' ) ;
                    state <= CACHE_MISS_WAIT_ONE;
                    mst_cntl_rd_req <= '1' ;
                    mst_cntl_wr_req <= '0' ;
                    open_operation <= '1';
                    match_address <= ( others => '0' ) ;
                    words_remaining <= conv_std_logic_vector (
                                2 ** CACHE_LINE_BITS , SP_LEFT + 1 ) ; 
                    trigger <= '1';
                    dbg_code <= x"25";
                    master_address ( 
                            CACHE_LINE_BITS + 2 - 1 downto 2 )
                            <= ( others => '0' ) ;
                    saved_be <= x"f" ;

            when CACHE_MISS_WAIT|CACHE_MISS_WAIT_ONE =>
                    dbg_code <= x"27";
                    if (( input_next = '1' )
                    and ( some_words_remain = '1' ))
                    then
                        dbg_code <= x"28";
                        sp_write <= '1' ;
                        bus_side_cen_override <= '1';
                        data_input_delay <= data_input ;
                        words_remaining <= 
                            words_remaining - 1 ;
                        state <= CACHE_MISS_WAIT;
                        if ( state = CACHE_MISS_WAIT )
                        then
                            bus_side_address <= bus_side_address_inc;
                        else
                            slow_data_out <= data_input ;
                            dbg_code <= x"2a";
                        end if;
                    end if ;
                    if (( input_next = '1' )
                    and ( state = CACHE_MISS_WAIT )
                    and ( bus_side_address_inc ( CACHE_LINE_BITS - 1 downto 0 ) 
                        = cache_saved_address ( CACHE_LINE_BITS - 1 downto 0 )))
                    then
                        slow_data_out <= data_input ;
                    end if ;

                    if ( busy = '0' )
                    then
                        -- Complete cache miss: send data to CPU
                        dbg_code <= x"29";
                        next_ack <= '1' ;
                        state <= READY ;
                    end if ;

            when BEGIN_OPEN =>
                    dbg_code <= x"00";
                    state <= HANDLE_TABLE_FULL;
                    if (saved_read_only = '1')
                    then
                        for i in 1 to NUM_READ_ONLY_ENTRIES
                        loop
                            if ( table(i).is_valid = '0' )
                            then
                                ref <= i;
                                state <= STAGE1_OPEN ;
                            end if;
                        end loop;
                    else
                        for i in NUM_READ_ONLY_ENTRIES + 1 to NUM_ENTRIES - 1
                        loop
                            if ( table(i).is_valid = '0' )
                            then
                                ref <= i;
                                state <= STAGE1_OPEN ;
                            end if;
                        end loop;
                    end if;
                    open_operation <= '1' ;
                    match_internal <= '1' ;

            when STAGE1_OPEN =>
                    table(ref).min_address <= match_address ( ACTIVE_LEFT downto 2 ) ;
                    table(ref).max_address <= 
                                match_address ( ACTIVE_LEFT downto 2 ) + 
                                words_remaining ;
                    table(ref).offset <= cpu_copy_loc -
                                        match_address ( SP_LEFT + 2 downto 2 ) ;
                    bus_side_address <= cpu_copy_loc - 1 ;
                    slow_data_out <= ( others => '0' ) ;
                    slow_data_out ( REF_LEFT downto 0 ) <= 
                        conv_std_logic_vector ( ref , REF_LEFT + 1 ) ;
                    state <= STAGE2_OPEN ;

            when STAGE2_OPEN =>
                    mst_cntl_rd_req <= '1' ;
                    master_address <= match_address ;
                    trigger <= '1' ;
                    bus_side_cen_override <= '1' ;
                    state <= STAGE3_OPEN ;

            when STAGE3_OPEN =>
                    if ( busy = '0' )
                    then
                        if ( some_words_remain = '1' )
                        then
                            state <= STAGE2_OPEN ;
                        else
                            table(ref).is_valid <= '1' ;
                            dbg_code <= x"09";
                            next_ack <= '1' ;
                            state <= READY ;
                        end if ;
                    end if ;

                    if (( input_next = '1' )
                    and ( some_words_remain = '1' ))
                    then
                        sp_write <= '1' ;
                        data_input_delay <= data_input ;
                        match_address <= match_address + 4 ;
                        bus_side_address <= bus_side_address_inc;
                        words_remaining <= words_remaining - 1 ;
                    end if ;
                    bus_side_cen_override <= '1' ;
            
            when BEGIN_CLOSE =>
                    dbg_code <= x"c0";
                    if ( table(ref).is_valid = '0' )
                    then
                        state <= HANDLE_NOT_FOUND ;
                    elsif ( ref <= NUM_READ_ONLY_ENTRIES )
                    then
                        -- Discard 
                        dbg_code <= x"c9";
                        table(ref).is_valid <= '0';
                        next_ack <= '1' ;
                        state <= READY ;
                    else
                        -- Write back
                        match_address <= ( others => '0' ) ;
                        match_address ( ACTIVE_LEFT downto 2 ) <= 
                                    table(ref).min_address;
                        words_remaining <= entry_size; 
                        bus_side_address <= close_start_address;
                        bus_side_cen_override <= '1' ;
                        close_operation <= '1' ;
                        state <= STAGE1_CLOSE ;
                    end if;
                    match_internal <= '1' ;
                    mst_cntl_wr_req <= '1' ;
      
            when STAGE1_CLOSE =>
                    master_address <= match_address ;

                    close_update_count <= 0;
                    read_only_update_pos <= 1;

                    row_activation <= row_validity;
                    for i in 2 to NUM_ENTRIES - 1
                    loop
                        if (( i <= NUM_READ_ONLY_ENTRIES )
                        or ( i >= ref ))
                        then
                            row_activation(i) <= '0';
                        end if;
                    end loop;
                    bus_side_cen_override <= '1' ;
                    read_only_copy_enable <= '1';
                    bus_side_address <= bus_side_address_inc;
                    state <= STAGE2_CLOSE ;

            when STAGE2_CLOSE =>
                    -- Fetch next address 
                    bus_side_address <= bus_side_address_inc;
                    close_update_count <= close_update_count + 1 ;
                    if ( close_null = '0' )
                    then
                        match_address <= match_address + 4 ;
                    end if;

                    if ( output_next = '1' )
                    then
                        if ( read_only_update_pos = 1 )
                        then
                            state <= HANDLE_DEADLINE_MISS;
                        else
                            words_remaining <= words_remaining - 1 ;
                        end if;
                    end if;

                    if (( close_update_count = INITIATE_POINT )
                    and ( read_only_update_pos = 1 ))
                    then
                        trigger <= '1';
                    end if;
                    
                    if ( close_update_count = BURST_SIZE - 1 )
                    then
                        row_activation <= ( others => '0' ) ;
                        if ( read_only_update_pos /= NUM_READ_ONLY_ENTRIES )
                        then
                            -- Update next read-only area
                            read_only_update_pos <= next_update_pos;
                            row_activation(next_update_pos) <= 
                                    row_validity(next_update_pos);
                            match_address <= ( others => '0' ) ;
                            match_address ( ACTIVE_LEFT downto 2 ) <= 
                                    table(ref).min_address;
                        else
                            state <= STAGE3A_CLOSE;
                            match_address <= match_address + 4 ;
                            read_only_copy_enable <= '0';
                        end if;
                    elsif ((( close_null = '1' )
                        and ( read_only_update_pos /= 1 ))
                    or ( close_update_count = BURST_SIZE - 2 ))
                    then
                        close_update_count <= BURST_SIZE - 1 ;
                        bus_side_address <= close_start_address;
                    end if ;
                    bus_side_cen_override <= '1' ;

            when STAGE3A_CLOSE =>
                    table(ref).min_address <= 
                            match_address ( ACTIVE_LEFT downto 2 ) ;
                    state <= STAGE3_CLOSE;
                    if ( output_next = '1' )
                    then
                        words_remaining <= words_remaining - 1 ;
                    end if;

            when STAGE3_CLOSE =>
                    if ( output_next = '1' )
                    then
                        words_remaining <= words_remaining - 1 ;
                    end if;
                    if ( busy = '0' )
                    then
                        if (some_words_remain = '0')
                        then
                            dbg_code <= x"c9";
                            table(ref).is_valid <= '0' ;
                            next_ack <= '1' ;
                            state <= READY ;
                        else
                            bus_side_address <= close_start_address;
                            state <= STAGE1_CLOSE ;
                            bus_side_cen_override <= '1' ;
                        end if;
                    end if ;

            when HANDLE_DEADLINE_MISS =>
                    dbg_code <= x"24";
                    if ( busy = '0' )
                    then
                        slow_data_out <= x"e0e0e014" ;
                        next_ack <= '1' ;
                        state <= READY ;
                    end if;

            when HANDLE_NOT_FOUND =>
                    dbg_code <= x"23";
                    slow_data_out <= x"e0e0e013" ;
                    next_ack <= '1' ;
                    state <= READY ;

            when HANDLE_TABLE_FULL =>
                    dbg_code <= x"20";
                    slow_data_out <= x"e0e0e010" ;
                    next_ack <= '1' ;
                    state <= READY ;

            when HANDLE_ERROR =>
                    dbg_code <= x"21";
                    if ( busy = '0' )
                    then
                        slow_data_out <= x"e0e0e011" ;
                        next_ack <= '1' ;
                        state <= READY ;
                    end if ;

            when HANDLE_TIMEOUT =>
                    if ( close_operation = '1' )
                    then
                        dbg_code <= x"c2";
                    elsif ( open_operation = '1' )
                    then
                        dbg_code <= x"02";
                    else
                        dbg_code <= x"22";
                    end if;

                    if ( busy = '0' )
                    then
                        slow_data_out <= x"e0e0e012" ;
                        next_ack <= '1' ;
                        state <= READY ;
                    end if ;

            end case ;

            -- Additional control lines
            if ( trigger = '1' )
            then
                mst_go_int <= '1' ;
            end if ;
            if ( mst_cmd_sm_clr_go = '1' )
            then
                mst_go_int <= '0' ;
            end if ;

            if ( mst_cmd_sm_set_timeout = '1' )
            then
                state <= HANDLE_TIMEOUT ;
            elsif ( mst_cmd_sm_set_error = '1' )
            then
                state <= HANDLE_ERROR ;
            end if ;

            if ( next_ack = '1' )
            then
                -- Activation registers: normal operation
                null;
            end if;
            if (( match_ro or match_rw ) = '1')
            then
                match_delay <= '1';
            end if;

            -- Table entry 0 can never be used.
            table(0).is_valid <= '0' ;
            table(0).min_address <= ( others => '0' ) ;
            table(0).max_address <= ( others => '0' ) ;
            table(0).offset <= ( others => '0' ) ;

            match_address ( 1 downto 0 ) <= "00" ;
            match_address ( CPU_LEFT downto ACTIVE_LEFT + 1 ) <= match_mask ;
        end if;
    end process;

    srg : for i in 0 to DATA_LEFT
    generate 
        srb : block
            signal shift_reg    : std_logic_vector(15 downto 0);
            signal ce           : std_logic;
        begin
            ce <= '1' when ((( state = STAGE2_CLOSE )
                    and ( read_only_update_pos = 1 ))
                    or ( output_next = '1' ))
                else '0';

            process ( clk ) is
            begin
                if ( clk'event and clk = '1' )
                then
                    if ( ce = '1' )
                    then
                        shift_reg <= shift_reg(14 downto 0) & 
                                        bus_side_data_out(i);
                    end if;
                end if;
            end process;
            shift_reg_out(i) <= shift_reg(15);
        end block srb ;
    end generate srg ; 



end architecture basic ;

