/*
 * Scratchpad MMU - Test program
 * Copyright (C) Jack Whitham 2009
 * http://www.jwhitham.org.uk/c/smmu.html
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation
 * (version 2.1 of the License only).
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include "xparameters.h"

#include "xuartns550_l.h"
#include "xutil.h"
#include "mb_interface.h"

#include "hwdriver.h"
#include "simif.h"


#define RNG_TABLE_SIZE      0x4000
#define BENCH_SIZE          2048

extern unsigned _binary_rngtable_bin_start[ RNG_TABLE_SIZE ] ;
static unsigned _rng_index_1 ;
static unsigned _rng_index_2 ;

void reset_rng ( void )
{
    _rng_index_1 = 54 ;
    _rng_index_2 = 1981 ;
}

unsigned rng ( void )
{
    unsigned i = 0;
    i = ( _binary_rngtable_bin_start[ _rng_index_1 ] >> 1 )
        ^ ( _binary_rngtable_bin_start[ _rng_index_2 ] >> 2 );

    _rng_index_1 ++;
    if ( _rng_index_1 >= RNG_TABLE_SIZE ) 
    {
        _rng_index_1 = 0 ;
        _rng_index_2 += 1337 ;
        _rng_index_2 %= RNG_TABLE_SIZE ;
    }
    return i ;
}

void printhex ( unsigned x )
{
    printf("%08x", x);
} 

void Test ( void ) ;
int braindamage_test ( void );
void access_benchmark(unsigned base_pointer, unsigned size_in_bytes);

extern unsigned ocresults [ 32 ] ;
extern void octest ( void ) ;

int main (void) 
{
    /*unsigned i ;*/
    unsigned area = (unsigned) _binary_rngtable_bin_start;
    int rc;

    //XUartNs550_SetBaud(XPAR_RS232_UART_1_BASEADDR, 
    //            XPAR_XUARTNS550_CLOCK_HZ, 115200);
    //XUartNs550_mSetLineControlReg(XPAR_RS232_UART_1_BASEADDR, 
    //            XUN_LCR_8_DATA_BITS);

    setvbuf ( stdout , NULL , _IONBF , 0 ) ;
    printf ( "\r\nSMMU Test Program for version 1.03\r\n" ) ;
    printf ( "GNU LGPL 2.1 free software, (C) Jack Whitham 2009-2010\r\n\r\n" );

    reset_rng () ;
#if XPAR_MICROBLAZE_0_USE_ICACHE
    microblaze_init_icache_range(0, XPAR_MICROBLAZE_0_CACHE_BYTE_SIZE);
    microblaze_enable_icache();
#endif

    /* Early test verifies halfword load/store */
    rc = braindamage_test ();
    if (rc != 0) {
        printf ( "The memory system is braindamaged: 0x%x\n", rc);
        return 0;
    }

    if ( HW_Probe () ) {
        HW_Init () ;
        Simulator_Init () ;

        printf ( "Your SMMU is believed to have %u " 
                "table entries, of which\r\n1 is a zero entry, and %u " 
                "are for read-only data.\r\nThe scratchpad size is %u " 
                "words (%u bytes). There is " ,
            NUM_ENTRIES, NUM_RO_ENTRIES, SP_SIZE, SP_SIZE * 4);

        if (HAS_CACHE) {
            printf ( "a dcache feature.\r\n"
                "Your dcache has %u pages, each containing %u lines. Each\r\n"
                "line has %u words. The page size is %u words (%u bytes).\r\n",
                    NUM_CACHE_PAGES, NUM_CACHE_LINES_PER_PAGE,
                    NUM_CACHE_WORDS_PER_LINE,
                    NUM_CACHE_LINES_PER_PAGE * NUM_CACHE_WORDS_PER_LINE,
                    NUM_CACHE_LINES_PER_PAGE * NUM_CACHE_WORDS_PER_LINE * 4);
        } else {
            printf ( "no dcache feature.\r\n" );
        }
        printf ( "\r\n" ) ;

        /* Mini memory benchmarks */
        {
            unsigned total, start, ref;

            access_benchmark(area, BENCH_SIZE);  /* warmup */
            start = CONTROL->global_counter;  
            access_benchmark(area, BENCH_SIZE);  /* for real */
            total = CONTROL->global_counter - start;  
            printf("Benchmark time (PLB via SMMU): %u cycles\n", total);

            ref = HW_OPEN (_binary_rngtable_bin_start, BENCH_SIZE, 0); 
            access_benchmark(area, BENCH_SIZE); /* warmup */
            start = CONTROL->global_counter;  
            access_benchmark(area, BENCH_SIZE); /* for real */
            total = CONTROL->global_counter - start;  
            HW_CLOSE(ref);
            printf ( "Benchmark time (Scratchpad via SMMU): "
                        "%u cycles\n", total);

            if (HAS_CACHE) {
                fflush(stdout);
                HW_Cache_Enable(1);
                access_benchmark(area, BENCH_SIZE); /* warmup */
                start = CONTROL->global_counter;
                access_benchmark(area, BENCH_SIZE); /* for real */
                total = CONTROL->global_counter - start;  
                HW_Cache_Disable();
                printf ( "Benchmark time (Scratchpad via "
                        "SMMU DC): %u cycles\n", total);
            }
        }

        /* Begin tests */
        Test () ; 
    } else {
        printf ( "SMMU not present...\n" ) ;
#define XPAR_MICROBLAZE_0_USE_DCACHE 1
#define XPAR_TMRCTR_0_BASEADDR 0x83c00000
#if XPAR_MICROBLAZE_0_USE_DCACHE
#if XPAR_TMRCTR_0_BASEADDR
        printf ( "but %u bytes of Microblaze dcache are present\n" ,
                XPAR_MICROBLAZE_0_DCACHE_BYTE_SIZE ) ;

        {
            unsigned total;
            unsigned start;
            volatile unsigned * timer = (unsigned *) XPAR_TMRCTR_0_BASEADDR;

            fflush(stdout);
            /* Enable timer in generate mode */
            timer[ 0 ] = 0x490;

            access_benchmark(area, BENCH_SIZE);  /* warmup */
            start = timer[ 2 ];
            access_benchmark(area, BENCH_SIZE);  /* for real */
            total = timer[ 2 ] - start;
            printf("Benchmark time (PLB direct): %u cycles\n", total);

            start = timer[ 2 ];
            access_benchmark(0, BENCH_SIZE);  
            total = timer[ 2 ] - start;
            printf("Benchmark time (LMB direct): %u cycles\n", total);

            /* Enable dcache */
            microblaze_init_dcache_range(0, XPAR_MICROBLAZE_0_DCACHE_BYTE_SIZE);
            microblaze_enable_dcache();
            access_benchmark(area, BENCH_SIZE);  /* warmup */
            start = timer[ 2 ];
            access_benchmark(area, BENCH_SIZE);  /* for real */
            total = timer[ 2 ] - start;
            printf("Benchmark time (Microblaze DC): %u cycles\n", total);
        }
#else
        printf("Timer not present.\n");
#endif
#else
        printf("Microblaze data cache not present.\n");
#endif
    }

    printf("The end. Be seeing you!\n");
    fflush(stdout);
    asm volatile ( ".long 0x80000001\n" ) ;
    return 0;
}

void __assert (const char * one, int two , const char * three)
{
    printf("\r\nBang! Assertion failed\r\n%s (%d) %s\r\n",
            one, two, three);
    asm volatile ( ".long 0x80000001\n" ) ;
    exit ( 1 ) ;
}



